#!/bin/sh
#
# gnrf.tcl - GnRythmFinder
#
# Author : Gilles Velay (November 1999)
#  email : gvelay@free.fr
#  url : http://gvelay.free.fr
#
# $Id: gnrf.tcl,v 1.8 2006/10/13 20:27:36 gilles Exp $
#
#\
exec wish "$0" "$@"  

#
# Global variables
#
global glob_counter_val_tab
global glob_counter_ms_tab
global glob_entry_tab
global glob_title_version

proc init {} {
    global glob_counter_val_tab
    global glob_counter_ms_tab
    global glob_title_version
    global glob_entry_tab

    # set title and version
    set glob_title_version "--GnRhythmFinder 1.2--"
    
    # reinit the table of counter clock values
    set glob_counter_val_tab {}

    # reinit the counter of ms
    set glob_counter_ms_tab {}

    # set default parameters
    set glob_entry_tab(nb_iter) 4
    set glob_entry_tab(nb_bars) 1
}

proc about_cb {} {
    global glob_title_version
    set msg    "$glob_title_version\n"
    append msg " (c) Gilles Velay, 2000-2006\n\n" 
    append msg "email : gvelay@free.fr\n"
    append msg "url:  http://gvelay.free.fr"
    tk_messageBox -title "About" -message $msg
}

proc manual_cb {} {
    set msg    "\tManual\n"
    append msg "\t------\n\n"
    append msg "1) Select \"Reset Recording\"\n\n"
    append msg "2) Put the mouse pointer on the Pad Zone\n\n"
    append msg "3) Play the riff many times in the Pad Zone by striking "
    append msg "any key\n\n"
    append msg "4) Fill in the parameters (nb of times you played the "
    append msg "riff, nb of bars in the riff and time signature)\n\n"
    append msg "5) Select \"Analyze\"\n\n"
    append msg "6) The rhythm is shown in a window :-)"
    tk_messageBox -title "How To use GnRythmFinder" -message $msg
}

proc pad_keypress_cb {} {
    global glob_counter_val_tab
    global glob_counter_ms_tab
    
    # record the counter clock value
    lappend glob_counter_val_tab [clock clicks]

    # record the counter ms value
    lappend glob_counter_ms_tab [clock clicks -millisecond]
}

proc reset_cb {} {
    global glob_counter_val_tab
    global glob_counter_ms_tab

    # reinit the table of counter clocks values
    set glob_counter_val_tab {}

    # reinit the counter of ms
    set  glob_counter_ms_tab {}
}

proc analyze_cb {} {
    global glob_counter_val_tab
    global glob_entry_tab
    global glob_counter_ms_tab

    set intervals {}
    set first 1
    
    # check that only one note was recorded
    if {[llength $glob_counter_val_tab] == 0} {
	tk_messageBox -message "No note was recorded\nPlease read the manual in the Help menu" -icon error
	return
    }

    # compute the time intervals
    foreach val $glob_counter_val_tab {
	if {!$first} {
	    lappend intervals [expr $val - $last_val]
	}
	set first 0
	set last_val $val
    }
	
    # compute conversion factor between clicks and seconds
    set first_click [lindex $glob_counter_val_tab 0]
    set last_click [lindex $glob_counter_val_tab end]
    set first_ms [lindex $glob_counter_ms_tab 0]
    set last_ms [lindex $glob_counter_ms_tab end]
    set clicks_per_ms [expr ($last_click - $first_click) / \
				($last_ms - $first_ms)]

    # Error Handling
    # Rk : the last note is recorded 1 time less than the others
    if {$glob_entry_tab(nb_bars) < 1} {
	tk_messageBox -message "Nb bars must be > 0" -icon error 
	return
    }
    if {$glob_entry_tab(nb_iter) < 1} {
	tk_messageBox -message "Nb iterations must be > 0" -icon error
	return
    }
    set nbNotes [expr [llength $intervals] + 1]
    if {[expr $nbNotes % $glob_entry_tab(nb_iter)] != 0} {
	set msg "Cannot compute average\n"
	set mod [expr $nbNotes % $glob_entry_tab(nb_iter)] 
	append msg "$mod note(s) to many"
	tk_messageBox -message $msg -icon error
	return
    }

    # compute average durations
    # -------------------------
    set nbNotesInRiff [expr $nbNotes / $glob_entry_tab(nb_iter)]
    set av_duration_tab {}
    
    # loop over each note (except last one)
    for {set indexNote 0} {$indexNote < [expr $nbNotesInRiff - 1]} \
	{incr indexNote} {
	    set duration [lindex $intervals $indexNote]
	    for {set indexRepeat 1} {$indexRepeat < $glob_entry_tab(nb_iter)} \
		{incr indexRepeat} {
		    set index [expr $indexRepeat * $nbNotesInRiff + $indexNote]
		    incr duration [lindex $intervals $index]
		}
	    lappend av_duration_tab [expr $duration / $glob_entry_tab(nb_iter)]
	}

    # special case for last note (played one time less)
    set index [expr $nbNotesInRiff - 1]
    set duration [lindex $intervals $index]
    for {set indexRepeat 1} \
	{$indexRepeat < [expr $glob_entry_tab(nb_iter)-1]} \
	{incr indexRepeat} {
	    set index [expr $indexRepeat * $nbNotesInRiff + ($nbNotesInRiff-1)]
	    incr duration [lindex $intervals $index]
	}
    lappend av_duration_tab [expr $duration / ($glob_entry_tab(nb_iter)-1)]

    # Sort out real durations
    # -----------------------
    
    # Normalization : 
    
    # compute total duration
    set total_duration 0
    foreach note $av_duration_tab { incr total_duration $note }
    
    # quarter notes are forced to have value 120
    set av_duration_norm_tab {}
    switch $glob_entry_tab(signature) {
	"4/4" { set factor [expr $total_duration / \
				(480*$glob_entry_tab(nb_bars))]
	}
	"5/4" { set factor [expr $total_duration / \
				(600*$glob_entry_tab(nb_bars))]
	}
	"7/4" {set factor [expr $total_duration / \
			       (840*$glob_entry_tab(nb_bars))]
	}
    }
    foreach duration $av_duration_tab {
	lappend av_duration_norm_tab [expr $duration / $factor]
    }
	
    # compute tempo

    set bar_time [expr $total_duration / $glob_entry_tab(nb_bars)]
    set bar_time [expr $bar_time / (60000.0 * $clicks_per_ms)]
    switch $glob_entry_tab(signature) {
	"4/4" { set tempo [expr 4 / $bar_time] 
	}
	"5/4" { set tempo [expr 5 / $bar_time] 
	}
	"7/4" { set tempo [expr 7 / $bar_time] 
	}
    }

    # Guess duration value of each note
    # ---------------------------------
    # w = 480
    # h = 240    h+q = 360    h+q+e = 420
    # q = 120    q3  = 160    q+e = 180
    # e = 60     e3  = 80     e+s = 90
    # s = 30     s3  = 40
    set riff {}
    foreach duration $av_duration_norm_tab {
	set note "?"
	if {$duration > 25 && $duration < 35} {
	    set note "s"
	} elseif {$duration < 50} {
	    set note "s3"
	} elseif {$duration < 70} {
	    set note "e"
	} elseif {$duration < 85} {
	    set note "e3"
	} elseif {$duration < 105} {
	    set note "e+s"
	} elseif {$duration < 140} {
	    set note "q"
	} elseif {$duration < 170} {
	    set note "q3"
	} elseif {$duration < 210} {
	    set note "q+e"
	} elseif {$duration < 300} {
	    set note "h"
	} elseif {$duration < 390} {
	    set note "h+q"
	} elseif {$duration < 450} {
	    set note "h+q+e" 
	} elseif {$duration >= 450} {
	    set note "w"
	}
	lappend riff $note
    }

    # Post Result
    # -----------
    set msg "tempo = $tempo\n\n"
    foreach note $riff {
	append msg "$note\n"
    }
    tk_messageBox -message $msg -title "Result" 
}

proc show_window {} {
    global glob_entry_tab
    global glob_title_version
    
    # create widgets
    # --------------
    wm title . $glob_title_version
    
    # menubar
    menu .menubar -type menubar
    menu .menubar.menuFile
    menu .menubar.menuHelp
    .menubar add cascade -label "File" -menu .menubar.menuFile
    .menubar add cascade -label "Help" -menu .menubar.menuHelp
    .menubar.menuFile add command -label "Quit" -command {exit}
    .menubar.menuHelp add command -label "Manual" -command {manual_cb}
    .menubar.menuHelp add command -label "About" -command {about_cb}
    . configure -menu .menubar

    # pad zone
    frame .frPadZone
    label .frPadZone.lab -text "Pad Zone"
    frame .frPadZone.frZone -height 100 -width 100 -background yellow
    bind .frPadZone.frZone <Enter> {
	focus %W
	%W configure -background blue
    }
    bind .frPadZone.frZone <Leave> {
	focus .
	%W configure -background yellow
    }
    bind .frPadZone.frZone <KeyPress> { 
	pad_keypress_cb 
    }

    # parameters
    frame .frParams
    frame .frParams.frIter
    label .frParams.frIter.lab -text "Nb of iterations :" -width 20
    entry .frParams.frIter.ent -textvariable glob_entry_tab(nb_iter) -width 2
    frame .frParams.frBars
    label .frParams.frBars.lab -text "Nb of bars :"  -width 20
    entry .frParams.frBars.ent -textvariable glob_entry_tab(nb_bars) -width 2
    frame .frParams.frSignature
    label .frParams.frSignature.lab -text "Signature :" -width 20 
    tk_optionMenu .frParams.frSignature.menu glob_entry_tab(signature) \
	"4/4" "5/4" "7/4"
                
    # buttons
    frame .frBut
    button .frBut.butReset -text "Reset Recording" -width 15 \
	-command {reset_cb}
    button .frBut.butAnalyze -text "Analyze" -width 15 \
	-command {analyze_cb}

    # set geometry
    # ------------
    pack .frPadZone.lab .frPadZone.frZone -side top
    pack .frParams.frIter.lab .frParams.frIter.ent -side left
    pack .frParams.frBars.lab .frParams.frBars.ent -side left
    pack .frParams.frSignature.lab .frParams.frSignature.menu -side left
    pack .frParams.frIter .frParams.frBars .frParams.frSignature -side top \
	-expand true -fill x  
    pack .frBut.butReset -side left
    pack .frBut.butAnalyze -side right
    # final pack
    pack .frPadZone .frParams .frBut -side top -pady 10 \
	-fill x -padx 16 
}

#------------------------------------------------------------
#
# script starts here
#
init
show_window
